/**
 * $Id: BagAsMapFactory.java,v 1.1 2006/02/07 19:30:05 uwe Exp $
 */
package de.fhw.data.bag.referenceImpl;

import de.fhw.data.map.MapFactory;
import de.fhw.data.map.referenceImpl.MapAsLinkedListFactory;
import de.fhw.data.bag.Bag;
import de.fhw.data.bag.BagFactory;

/**
 * A concrete factory for bag implementations via maps.
 * The attribute value of the map elements is the # of occurences.
 * 
 * @param <E> the element type
 */

public
class BagAsMapFactory<E> implements BagFactory<E> {
	final MapFactory<E,Integer> mf;
	/**
	 * 
	 * @param mf the factory for the underlying map implementation.
	 * If this factory is yields an efficient implementation,
	 * the bag implementation also becomes rather efficient
	 */
	public BagAsMapFactory(MapFactory<E,Integer> mf) {
		this.mf = mf;
	}
	/**
	 * the default map factory is for the reference implementation
	 * in de.fhw.data.map.referenceImpl, a very inefficient one
	 *
	 */
	public BagAsMapFactory() {
		this(new MapAsLinkedListFactory<E,Integer>());
	}
	/**
	 * generation of an empty bag. The result can be used as a prototype.
	 */
	public Bag<E> empty() {
		return
		new BagAsMapImpl<E>(mf);
	}
}
