/**
 * $Id: ListAsLinkedList.java,v 1.3 2006/05/15 11:23:08 uwe Exp $
 */
package de.fhw.data.list.referenceImpl;

import de.fhw.data.function.Fct;
import de.fhw.data.function.Fct2;
import de.fhw.data.list.List;
import de.fhw.data.list.ListImpl;
import de.fhw.data.tuple.Pair;

/**
 * minimal implementation for lists as single linked lists
 *
 */
public
class ListAsLinkedList<E>
extends ListImpl<E> {

	private static
	class Node <E> {
		E e; 
		Node <E> next;
		
		protected Node(E e, Node <E> l) {
			this.e = e;
			this.next = l;
			++newNode;
		}
	}
	
	final Node<E> data;
	
	ListAsLinkedList () {
		this((Node<E>)null);
	}
	
	private
	ListAsLinkedList(Node<E> n) {
		data = n;
		++newList;
	}
		
	//--------------------
	
	public
	List<E> cons(E e)
	{
		return
		new ListAsLinkedList<E>(
				new Node<E> (e,data)
				);
	}
	
	@Override
	public
	boolean isEmpty() {
		return
		data == null;
	}
	
	public
	List <E> empty() {
		if (isEmpty())
			return this;
		return
		new ListAsLinkedList<E>();
	}
	
	@Override
	public
	int length() {
		int res = 0;
		Node<E>l = data;
		while (l != null) {
			++res;
			l = l.next;
		}
		return res;
	}

	@Override
	public
	E head() {
		if (data == null)
			throw new IndexOutOfBoundsException("head applied to empty list");
		
		return data.e;
	}

	@Override
	public
	List <E> tail() {
		if (data == null)
			throw new IndexOutOfBoundsException("tail applied to empty list");
		
		return new ListAsLinkedList<E>(data.next);
	}

	@Override
	public
	E last() {
		if (data == null)
			throw new IndexOutOfBoundsException("last applied to empty list");
		Node<E> l = data;
		while (l.next != null) {
			l = l.next;
		}
		return l.e;
	}

	@Override
	public
	List<E> take(int i) {
		return
		new ListAsLinkedList<E>(takeN(data,i));
	}

	private
	Node<E> takeN(Node<E> l, int i) {
		if (l == null || i <= 0)
			return null;
		return
		new Node<E>(l.e, takeN(l.next,i-1));
	}
	
	@Override
	public
	List<E> drop(int i) {
		Node<E> l = data;
		while (l.next != null && i > 0) {
			++i;
			l = l.next;
		}
		return new ListAsLinkedList<E>(l);
	}

	@Override
	public
	List<E> concat(final List<E> l2) {
		if (isEmpty())
			return l2;

		if (l2.isEmpty())
			return this;
		
		return
		new ListAsLinkedList<E>(
				concatN(data,
						((ListAsLinkedList<E>)l2).data));
	}

	private
	Node<E> concatN(Node<E>l1, Node<E>l2) {
		if (l1 == null)
			return l2;
		
		return
		new Node<E>(l1.e,concatN(l1.next,l2));
	}
	
	public
	<R>
	R foldl(final Fct2<R,E,R> f, R init) {
		return
		foldlN(data, f, init);
	}

	private
	<R>
	R foldlN(
			Node<E> l,
			final Fct2<R,E,R> f, R acc) {
		while (l != null) {
			acc = f.apply(acc,l.e);
			l = l.next;
		}
		return acc;
	}
		
	public
	<R>
	R foldr(final Fct2<E,R,R> f, R init) {
		return
		foldrN(data, f, init);
	}
	
	private
	<R>
	R foldrN(
			Node<E> l,
			final Fct2<E,R,R> f, R acc) {
		return
		(l == null)
		? acc : f.apply(l.e, foldrN(l.next,f,acc));
	}

	public
	<R>
	R foldrMap(
			final Fct2<R,R,R> cf,
			final Fct<E,R> f,
			R init) {
		Fct2<E,R,R> foldf =
			new Fct2<E,R,R>() {
			public
			R apply(E e, R r) {
				return
				cf.apply(f.apply(e),r);
			}
		};
		
		return
		foldrN(data, foldf, init);
	}
	
	public
	<R>
	R foldr1Map(
			final Fct2<R,R,R> combine, final Fct<E,R> map, R emptyRes) {
		return
		(data == null)
		? emptyRes
				: foldrN1Map(data, combine, map);
	}
	
	private
	<R>
	R foldrN1Map(
			Node<E> l,
			final Fct2<R,R,R> combine,
			final Fct<E,R> map) {
		return
		(l.next == null)
		? map.apply(l.e)
				: combine.apply(map.apply(l.e), foldrN1Map(l.next, combine, map));
	}
	
	public
	Pair<Integer, Integer> objectsUsed() {
		Node<E> l = data;
		int len = 0;
		
		while (l != null) {
			++len;
			l = l.next;
		}
		
		int noOfObjects = len + 1;	// l nodes, 1 list
		int noOfFields  = 2 * len + 1;	// 2 * l node fields, 1 list field
		return
		new Pair<Integer, Integer>(noOfObjects, noOfFields);
	}
	
	/**
	 * the static counters for constructor calls for the wrapper and the node objects
	 */
	static int newList  = 0;
	static int newNode = 0;
	
	public
	Pair<Integer, Integer> objectsAllocated() {
		int nm = newList;
		int nn = newNode;
		newList  = 0;
		newNode = 0;
		return
		new Pair<Integer,Integer>(nm + nn, nm + 2 * nn);
	}
	
	public
	String fingerprint() {
		Node<E> l = data;
		StringBuffer res = new StringBuffer();
		res.append(this.hashCode());
		
		while (l != null) {
			res.append(",").append(l.hashCode());
			res.append(",").append(l.e.hashCode());
			l = l.next;
		}
		return
		res.toString();
	}
}
