/**
 * $Id: MapAsLinkedList.java,v 1.6 2006/02/24 21:21:02 uwe Exp $
 */

package de.fhw.data.map.referenceImpl;

import de.fhw.data.function.Fct;
import de.fhw.data.function.Fct2;
import de.fhw.data.function.Fct3;
import de.fhw.data.map.Map;
import de.fhw.data.map.MapImpl;
import de.fhw.data.tuple.Pair;

/**
 * Reference implementation for maps.
 * This implementation supports the minimal
 * set of methods neccessary for a concrete implementation.
 * 
 * Implementation strategie is a linked list like in C
 * The list is represented by a wrapper object. This object contains
 * the key value pairs as a list of nodes
 * 
 * @author uwe
 *
 * @param <K> the type of the Key
 * @param <V> the type of the attribute
 */
class MapAsLinkedList<K,V> extends MapImpl<K,V> {
	
	/**
	 * The auxiliary data class for the nodes of a list
	 * 
	 * @author uwe
	 *
	 * @param <K> the type of the Key
	 * @param <V> the type of the Value
	 */
	private static
	class Node <K,V> {
		K k;
		V v; 
		Node <K,V> next;
		
		protected Node(K k, V v, Node <K,V> l) {
			this.k = k;
			this.v = v;
			this.next = l;
			++newNode;
		}
	}
	
	final Node<K,V> data;
	
	MapAsLinkedList () {
		this((Node<K,V>)null);
	}
	
	private
	MapAsLinkedList(Node<K,V> n) {
		data = n;
		++newMap;
	}
		
	//--------------------
	
	private
	Map <K,V> cons(K k, V v)
	{
		return
		new MapAsLinkedList<K,V>(
				new Node<K,V> (k,v,data)
				);
	}
	
	@Override
	public
	boolean isEmpty() {
		return
		data == null;
	}
	
	public
	Map <K,V> empty() {
		if (isEmpty())
			return this;
		return
		new MapAsLinkedList<K,V>();
	}
	
	public
	Map <K,V> insert(K k, V v) {
		if (! contains(k))
			return
			cons(k,v);
		return
		remove(k).insert(k,v);
	}
	
	@Override
	protected
	Pair <K,V> choose1() {
		if (data == null)
			return (Pair<K,V>)null;
		return
		new Pair<K,V>(data.k,data.v);
	}
	
	@Override
	public
	<V2>
	Map <K,V2> map(
			final Fct<V, V2> mf,
			Map <K,V2> proto) {
		
		Fct3<K,V,Map<K,V2>,Map<K,V2>> fo
		= new Fct3<K,V,Map<K,V2>,Map<K,V2>>(){
			public
			Map <K,V2> apply(
					K k,
					V v,
					Map <K,V2> r) {
				return
				r.insert(k,mf.apply(v));
			}
		};
		return
		foldN(data,fo,proto.empty());
	}
	
	private
	<V2>
	Node<K,V2> mapN(
			Node<K,V> l,
			final Fct<V,V2> mf) {
		
		return
		(l == null)
		? (Node<K,V2>)null
				: new Node<K,V2>(l.k, mf.apply(l.v), mapN(l.next, mf));
	}
	
	public
	<R>
	R fold(final Fct3<K,V,R,R> f, R init) {
		return
		foldN(data, f, init);
	}
	
	private
	<R>
	R foldN(
			Node<K,V> l,
			final Fct3<K,V,R,R> f, R acc) {
		return
		(l == null)
		? acc
				: foldN(l.next, f, f.apply(l.k, l.v, acc));
	}

	public
	<R>
	R foldMap(
			final Fct2<R,R,R> cf,
			final Fct2<K,V,R> f,
			R init) {
		Fct3<K,V,R,R> foldf =
			new Fct3<K,V,R,R>() {
			public
			R apply(K k, V v, R r) {
				return
				cf.apply(f.apply(k,v),r);
			}
		};
		
		return
		foldN(data, foldf, init);
	}
	
	public
	<R>
	R fold1Map(
			final Fct2<R,R,R> cf, final Fct2<K,V,R> f, R emptyRes) {
		return
		(data == null)
		? emptyRes
				: foldN1Map(data, cf, f);
	}
	
	private
	<R>
	R foldN1Map(
			Node<K,V> l,
			final Fct2<R,R,R> cf,
			final Fct2<K,V,R> f) {
		return
		(l.next == null)
		? f.apply(l.k, l.v)
				: cf.apply(f.apply(l.k, l.v), foldN1Map(l.next, cf, f));
	}
	
	public
	boolean inv() {
		return
		true;
	}
	
	public
	Pair<Integer, Integer> objectsUsed() {
		Node<K,V> l = data;
		int len = 0;
		
		while (l != null) {
			++len;
			l = l.next;
		}
		
		int noOfObjects = len + 1;	// l nodes, 1 list
		int noOfFields  = 3 * len + 1;	// 3 * l node fields, 1 list field
		return
		new Pair<Integer, Integer>(noOfObjects, noOfFields);
	}
	/**
	 * the static counters for construtor calls for the wrapper and the node objects
	 */
	static int newMap  = 0;
	static int newNode = 0;
	
	public
	Pair<Integer, Integer> objectsAllocated() {
		int nm = newMap;
		int nn = newNode;
		newMap  = 0;
		newNode = 0;
		return
		new Pair<Integer,Integer>(nm + nn, nm + 3 * nn);
	}

	public
	String fingerprint() {
		Node<K,V> l = data;
		StringBuffer res = new StringBuffer();
		res.append(this.hashCode());
		
		while (l != null) {
			res.append(",").append(l.hashCode());
			res.append(",").append(l.k.hashCode());
			res.append(",").append(l.v.hashCode());
			l = l.next;
		}
		return
		res.toString();
	}
	
}
