package de.fhw.data.set.referenceImpl;

import de.fhw.data.function.Fct;
import de.fhw.data.function.Fct2;
import de.fhw.data.function.Fct3;
import de.fhw.data.map.Map;
import de.fhw.data.map.MapFactory;
import de.fhw.data.set.Set;
import de.fhw.data.set.SetImpl;
import de.fhw.data.tuple.Pair;

public class SetAsMapImpl<E> extends SetImpl<E> {
	private static final Integer v = 42;
	
	final MapFactory<E,Integer> mf;
	final Map<E,Integer> m;
	
	SetAsMapImpl(MapFactory<E,Integer> mf, Map<E,Integer> m) {
		this.mf = mf;
		this.m  = m;
		++newSet;
	}
	
	SetAsMapImpl(MapFactory<E,Integer> mf) {
		this(mf,mf.empty());
	}
	/**
	 * generate a new empty set
	 */
	public Set<E> empty() {
		if (isEmpty())
			return this;
		return
		new SetAsMapImpl<E>(mf);
	}
	/**
	 * reimplement isEmpty for efficiency
	 */
	@Override
	public boolean isEmpty() {
		return
		m.isEmpty();
	}
	/**
	 * insert implemented via map insert
	 */
	public Set<E> insert(E e) {
		if (m.contains(e))
			return this;
		return
		new SetAsMapImpl<E>(mf,m.insert(e,v));
	}
	/**
	 * remove implemented via map remove
	 */
	@Override
	public Set<E> remove(E e) {
		if (! m.contains(e))
			return this;
		return
		new SetAsMapImpl<E>(mf,m.remove(e));
	}
	
	/**
	 * implement choose1 via map choose
	 */
	@Override
	protected
	E choose1() {
		Pair<E,Integer> e = m.choose();
		return
		e == null
		? null : e.fst;
		
	}
	/**
	 * implementation of fold via map fold
	 */
	// @Override
	public
	<R>
	R fold(final Fct2<E,R,R> f, R init) {
		Fct3<E,Integer,R,R> fo
		= new Fct3<E,Integer,R,R>() {
			public
			R apply(E k, Integer v, R r) {
				return
				f.apply(k,r);
			}
		};
		return m.fold(fo,init);
	}

	// TODO this declaration leads to error: method becomes ambiguous
	/**
	 * Ecipse bug: The method is ambiguous for the type ...
	 * 
	 * If bug is removed, delete this wrapper method
	 */

	public
	<R>
	R foldMap(
			final Fct2<R,R,R> combine,
			final Fct<E,R> map,
			R init) {
		return eclipseBugFoldMap(combine,map,init);
	}

	// TODO this declaration leads to error: method becomes ambiguous
	/**
	 * Ecipse bug: The method is ambiguous for the type ...
	 * 
	 * If bug is removed, delete this wrapper method
	 */
	public
	<R>
	R fold1Map(
			final Fct2<R,R,R> combine,
			final Fct<E,R> map,
			R emptyRes) {
		return eclipseBugFold1Map(combine,map,emptyRes);
	}
	/**
	 * implemented by map invariant
	 */
	public boolean inv() {
		return m.inv();
	}
	/**
	 * add map space and wrapper object space
	 */
	public Pair<Integer, Integer> objectsUsed() {
		Pair<Integer,Integer> msp = m.objectsUsed();
		
		int noOfObjects = msp.fst + 1;	// m nodes, 1 wrapper
		int noOfFields  = msp.snd + 2;	// n node fields, 2 wrapper field
		return
		new Pair<Integer, Integer>(noOfObjects, noOfFields);
	}
	
	/**
	 * the static counters for construtor calls for the wrapper and the node objects
	 */
	static int newSet  = 0;
	
	public
	Pair<Integer, Integer> objectsAllocated() {
		int nm = newSet;
		newSet = 0;
		Pair<Integer,Integer> sp = m.objectsAllocated();
		
		return
		new Pair<Integer,Integer>(nm + sp.fst, 2 * nm + sp.snd);
	}
	
	public
	String fingerprint() {
		StringBuffer res = new StringBuffer();
		res.append(this.hashCode());
		res.append(",").append(mf.hashCode());
		res.append(",").append(m.fingerprint());

		return
		res.toString();
	}	
}
